/* Centaur Terminal File Client
 *
 * Version 1.0
 * 1/8/2021
 *
 * ctfc.c
 */

#include <arpa/inet.h>
#include <ctype.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#define PROG_HELP \
"Centaur Terminal File Client: Download and read files from terminal file servers\n\
\n\
ctfc [-H] [-d] <host> <port> <file> [output-file]\n\
\n\
\t[-H] - Display program syntax and options\n\
\t[-d] - Print output to standard output; do not download\n\
\n\
\t<host> must be a valid domain name or IP address.\n\
\t<port> must be a valid TCP port number.\n\
\t<file> must be a file listed in the server manifest, or '_' for the manifest itself.\n\
\t[output-file] can be specified to download files that exist in subdirectories without\n\
\tcreating them on the local machine.\n"

#define DEFAULT_PORT	4444

enum
{
	_NONE,
	_FILE,
	_DIR
};

int ftype(char file[])
{
	struct stat fstat;
	int s = stat(file, &fstat);
	if (s != 0)
	{
		return _NONE;
	}
	if (S_ISREG (fstat.st_mode)) 
	{
		return _FILE;
	} else if (S_ISDIR (fstat.st_mode)) 
	{
		return _DIR;
	} else {
		return _NONE;
	}
}

char * get_ip(char hostname[])
{
	static char ip[16];
	struct hostent *ent;
	struct in_addr **addr;

	if ((ent = gethostbyname(hostname)) == NULL)
	{
		return "(not found)";
	}

	addr = (struct in_addr **) ent->h_addr_list;

	strcpy(ip, inet_ntoa(*addr[0]));

	return ip;
}

int main(int argc, char *argv[])
{
	int no_dl = 0;
	int no_out = 0;

	int a_host = 1;
	int a_port = 2;
	int a_file = 3;
	int a_ofil = 4;
	
	char ofile[64];

	if (argc > 1 && argv[1][0] == '-' && argv[1][1] == 'H')
	{
		printf(PROG_HELP);
		exit(0);
	}

	if (argc > 1 && argv[1][0] == '-' && argv[1][1] == 'd')
	{
		no_dl = 1;
	}

	if (no_dl)
	{
		a_host++;
		a_port++;
		a_file++;
		a_ofil++;
	}

	if ((!no_dl && argc < 4) || (no_dl && argc < 5))
	{
		printf("Usage: %s [-H] [-d] <host> <port> <file> [output-file]\n", argv[0]);
		exit(1);
	}

	if ((!no_dl && argc < 5) || (no_dl && argc < 6))
	{
		strcpy(ofile, argv[a_file]);
	} else {
		no_out = 1;
		strcpy(ofile, argv[a_ofil]);
	}

	int sock;
	int fd;
	int rd;
	int c;
	struct sockaddr_in server;
	char buffer[102400];

	if (ftype(argv[a_file]) != _NONE && !no_dl && !no_out)
	{
		printf("%s: local file '%s' already exists\n", argv[0], argv[a_file]);
		exit(1);
	}

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == 0)
	{
		printf("%s: failed to create socket\n", argv[0]);
		exit(1);
	}

	server.sin_family = AF_INET;
	server.sin_port = htons(atoi(argv[a_port]));

	if ((inet_pton(AF_INET, get_ip(argv[a_host]), &server.sin_addr)) < 0)
	{
		printf("%s: address not supported\n", argv[0]);
		exit(1);
	}

	if ((connect(sock, (struct sockaddr *)&server, sizeof(server))) < 0)
	{
		printf("%s: connection failed\n", argv[0]);
		exit(1);
	}

	if (!strcmp(argv[a_file], "_"))
	{
		no_dl = 1;
		send(sock, "MANIFEST", 8, 0);
	}

	send(sock, argv[a_file], strlen(argv[a_file]), 0);
	
	if (!no_dl)
	{
		fd = open(ofile, O_WRONLY|O_CREAT, 0644);
		if (fd < 0)
		{
			fprintf(stderr, "%s: failed to create '%s'\nDirectory (if specified) must be present on local machine\n", argv[0], argv[a_file]);
			exit(1);
		}
		close(1);
		dup(fd);
	}

	while (1)
	{
		rd = read(sock, buffer, 102400);

		if (rd <= 0)
		{
			break;
		}

		for (c = 0; c < rd; c++)
		{
			putchar(buffer[c]);
		}
		
	}

	if (!no_dl)
	{
		close(fd);
	}
	close(sock);

	return 0;
}
